@@ -1,29 +1,30 @@
-##############################################################################
-##                                                                          ##
-## snimpy -- Interactive SNMP tool                                          ##
-##                                                                          ##
-## Copyright (C) Vincent Bernat <bernat@luffy.cx>                           ##
-##                                                                          ##
-## Permission to use, copy, modify, and distribute this software for any    ##
-## purpose with or without fee is hereby granted, provided that the above   ##
-## copyright notice and this permission notice appear in all copies.        ##
-##                                                                          ##
-## THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES ##
-## WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF         ##
-## MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR  ##
-## ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES   ##
-## WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN    ##
-## ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF  ##
-## OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.           ##
-##                                                                          ##
-##############################################################################
+#
+# snimpy -- Interactive SNMP tool
+#
+# Copyright (C) Vincent Bernat <bernat@luffy.cx>
+#
+# Permission to use, copy, modify, and/or distribute this software for any
+# purpose with or without fee is hereby granted, provided that the above
+# copyright notice and this permission notice appear in all copies.
+#
+# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
+# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
+# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
+# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
+# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
+# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
+# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
+#
 
-"""
-simple interface to libsmi
-"""
+"""This module is a low-level interface to manipulate and
+extract information from MIB files. It is a CFFI_ wrapper around
+libsmi_. You may find convenient to use it in other projects but the
+wrapper is merely here to serve *Snimpy* use and is therefore
+incomplete.
 
-from __future__ import print_function
-from __future__ import unicode_literals
+.. _libsmi: http://www.ibr.cs.tu-bs.de/projects/libsmi/
+.. _CFFI: http://cffi.readthedocs.org/
+"""
 
 from cffi import FFI
 
@@ -126,6 +127,7 @@ char        *smiLoadModule(const char *);
 SmiModule   *smiGetModule(const char *);
 SmiModule   *smiGetNodeModule(SmiNode *);
 SmiType     *smiGetNodeType(SmiNode *);
+SmiType     *smiGetParentType(SmiType *);
 char        *smiRenderNode(SmiNode *, int);
 SmiElement  *smiGetFirstElement(SmiNode *);
 SmiElement  *smiGetNextElement(SmiElement *);
@@ -149,75 +151,105 @@ SmiNamedNumber *smiGetNextNamedNumber(SmiNamedNumber *);
 
 _smi = ffi.verify("""
 #include <smi.h>
-""", libraries=[b"smi"])
+""", libraries=["smi"])
+
 
 class SMIException(Exception):
-    """SMI related exception"""
 
-class Entity(object):
-    """MIB entity"""
+    """SMI related exception. Any exception thrown in this module is
+    inherited from this one."""
+
+
+class Node(object):
+
+    """MIB node. An instance of this class represents a MIB node. It
+    can be specialized by other classes, like :class:`Scalar`,
+    :class:`Table`, :class:`Column`, :class:`Node`.
+    """
 
     def __init__(self, node):
+        """Create a new MIB node.
+
+        :param node: libsmi node supporting this node.
+        """
         self.node = node
 
     @property
     def type(self):
-        """Get the basic type associated with this entity.
+        """Get the basic type associated with this node.
 
-        @return: class from basictypes module which can represent the entity.
+        :return: The class from :mod:`basictypes` module which can
+            represent the node. When retrieving a valid value for
+            this node, the returned class can be instanciated to get
+            an appropriate representation.
         """
         from snimpy import basictypes
         t = _smi.smiGetNodeType(self.node)
         target = {
             _smi.SMI_BASETYPE_INTEGER32: basictypes.Integer,
             _smi.SMI_BASETYPE_INTEGER64: basictypes.Integer,
-            _smi.SMI_BASETYPE_UNSIGNED32: { "TimeTicks": basictypes.Timeticks,
-                                            None: basictypes.Unsigned32 },
+            _smi.SMI_BASETYPE_UNSIGNED32: {b"TimeTicks": basictypes.Timeticks,
+                                           None: basictypes.Unsigned32},
             _smi.SMI_BASETYPE_UNSIGNED64: basictypes.Unsigned64,
-            _smi.SMI_BASETYPE_OCTETSTRING: { "IpAddress": basictypes.IpAddress,
-                                             "MacAddress": basictypes.MacAddress,
-                                             None: basictypes.String },
+            _smi.SMI_BASETYPE_OCTETSTRING: {b"IpAddress": basictypes.IpAddress,
+                                            None: basictypes.OctetString},
             _smi.SMI_BASETYPE_OBJECTIDENTIFIER: basictypes.Oid,
-            _smi.SMI_BASETYPE_ENUM: { "TruthValue": basictypes.Boolean,
-                                      None: basictypes.Enum },
+            _smi.SMI_BASETYPE_ENUM: {b"TruthValue": basictypes.Boolean,
+                                     None: basictypes.Enum},
             _smi.SMI_BASETYPE_BITS: basictypes.Bits
         }.get(t.basetype, None)
         if isinstance(target, dict):
-            target = target.get(t.name != ffi.NULL
-                                and ffi.string(t.name)
-                                or None, target.get(None, None))
+            tt = _smi.smiGetParentType(t)
+            target = target.get((t.name != ffi.NULL and ffi.string(t.name)) or
+                                (tt.name != ffi.NULL and ffi.string(
+                                    tt.name)) or None,
+                                target.get(None, None))
+
         if target is None:
-            raise SMIException("unable to retrieve type of entity")
+            raise SMIException("unable to retrieve type of node")
         return target
 
     @property
     def fmt(self):
-        """Get entity format
+        """Get node format. The node format is a string to use to display
+        a user-friendly version of the node. This is can be used for
+        both octet strings or integers (to make them appear as decimal
+        numbers).
+
+        :return: The node format as a string or None if there is no
+            format available.
 
-        @return: entity format or None if there is none
         """
         t = _smi.smiGetNodeType(self.node)
-        if t == ffi.NULL or t.format == ffi.NULL:
+        tt = _smi.smiGetParentType(t)
+        f = (t != ffi.NULL and t.format != ffi.NULL and ffi.string(t.format) or
+             tt != ffi.NULL and tt.format != ffi.NULL and
+             ffi.string(tt.format)) or None
+        if f is None:
             return None
-        return ffi.string(t.format)
+        return f.decode("ascii")
 
     @property
     def oid(self):
-        """Get OID for the current entity
+        """Get OID for the current node. The OID can then be used to request
+        the node from an SNMP agent.
 
-        @return: OID as a tuple
+        :return: OID as a tuple
         """
         return tuple([self.node.oid[i] for i in range(self.node.oidlen)])
 
     @property
     def ranges(self):
-        """Get entity ranges.
+        """Get node ranges. An node can be restricted by a set of
+        ranges. For example, an integer can only be provided between
+        two values. For strings, the restriction is on the length of
+        the string.
 
-        The returned value can be C{None} if no restriction on range
-        exists for the current entity, a single value if the range is
+        The returned value can be `None` if no restriction on range
+        exists for the current node, a single value if the range is
         fixed or a list of tuples or fixed values otherwise.
 
-        @return: valid range for the entity.
+        :return: The valid range for this node.
         """
         t = _smi.smiGetNodeType(self.node)
         if t == ffi.NULL:
@@ -231,7 +263,7 @@ class Entity(object):
             if m1 == m2:
                 ranges.append(m1)
             else:
-                ranges.append((m1,m2))
+                ranges.append((m1, m2))
             range = _smi.smiGetNextRange(range)
         if len(ranges) == 0:
             return None
@@ -241,9 +273,11 @@ class Entity(object):
 
     @property
     def enum(self):
-        """Get possible enum values.
+        """Get possible enum values. When the node can only take a discrete
+        number of values, those values are defined in the MIB and can
+        be retrieved through this property.
 
-        @return: dictionary of possible values keyed by the integer value.
+        :return: The dictionary of possible values keyed by the integer value.
         """
         t = _smi.smiGetNodeType(self.node)
         if t == ffi.NULL or t.basetype not in (_smi.SMI_BASETYPE_ENUM,
@@ -253,18 +287,19 @@ class Entity(object):
         result = {}
         element = _smi.smiGetFirstNamedNumber(t)
         while element != ffi.NULL:
-            result[self._convert(element.value)] = ffi.string(element.name)
-            element = _smi.smiGetNextNamedNumber(element);
+            result[self._convert(element.value)] = ffi.string(
+                element.name).decode("ascii")
+            element = _smi.smiGetNextNamedNumber(element)
         return result
 
     def __str__(self):
-        return ffi.string(self.node.name)
+        return ffi.string(self.node.name).decode("ascii")
 
     def __repr__(self):
         r = _smi.smiRenderNode(self.node, _smi.SMI_RENDER_ALL)
         if r == ffi.NULL:
-            return "<uninitialized {0} object at {1}>".format(self.__class__.__name__,
-                                                              hex(id(self)))
+            return "<uninitialized {0} object at {1}>".format(
+                self.__class__.__name__, hex(id(self)))
         module = _smi.smiGetNodeModule(self.node)
         if module == ffi.NULL:
             raise SMIException("unable to get module for {0}".format(
@@ -274,25 +309,38 @@ class Entity(object):
                                              ffi.string(module.name))
 
     def _convert(self, value):
-        attr = { _smi.SMI_BASETYPE_INTEGER32: "integer32",
-                 _smi.SMI_BASETYPE_UNSIGNED32: "unsigned32",
-                 _smi.SMI_BASETYPE_INTEGER64: "integer64",
-                 _smi.SMI_BASETYPE_UNSIGNED64: "unsigned64" }.get(value.basetype, None)
+        attr = {_smi.SMI_BASETYPE_INTEGER32: "integer32",
+                _smi.SMI_BASETYPE_UNSIGNED32: "unsigned32",
+                _smi.SMI_BASETYPE_INTEGER64: "integer64",
+                _smi.SMI_BASETYPE_UNSIGNED64: "unsigned64"}.get(value.basetype,
+                                                                None)
         if attr is None:
             raise SMIException("unexpected type found in range")
         return getattr(value.value, attr)
 
-class Scalar(Entity):
-    """MIB scalar entity"""
 
-class Table(Entity):
-    """MIB table entity"""
+class Scalar(Node):
+
+    """MIB scalar node. This class represents a scalar value in the
+    MIB. A scalar value is a value not contained in a table.
+    """
+
+
+class Table(Node):
+
+    """MIB table node. This class represents a table. A table is an
+    ordered collection of objects consisting of zero or more
+    rows. Each object in the table is identified using an index. An
+    index can be a single value or a list of values.
+    """
 
     @property
     def columns(self):
-        """Get table columns.
+        """Get table columns. The columns are the different kind of objects
+        that can be retrieved in a table.
+
+        :return: list of table columns (:class:`Column` instances)
 
-        @return: list of table columns
         """
         child = _smi.smiGetFirstChildNode(self.node)
         if child == ffi.NULL:
@@ -314,16 +362,17 @@ class Table(Entity):
 
     @property
     def _row(self):
-        """Get table row.
+        """Get the table row.
 
-        @return: row object
+        :return: row object (as an opaque object)
         """
         child = _smi.smiGetFirstChildNode(self.node)
         if child != ffi.NULL and child.indexkind == _smi.SMI_INDEX_AUGMENT:
             child = _smi.smiGetRelatedNode(child)
             if child == ffi.NULL:
-                raise SMIException("AUGMENT index for {0} but unable to retrieve it".format(
-                    ffi.string(self.node.name)))
+                raise SMIException("AUGMENT index for {0} but "
+                                   "unable to retrieve it".format(
+                                       ffi.string(self.node.name)))
         if child == ffi.NULL:
             raise SMIException("{0} does not have a row".format(
                 ffi.string(self.node.name)))
@@ -332,16 +381,19 @@ class Table(Entity):
                 ffi.string(child.name),
                 ffi.string(self.node.name)))
         if child.indexkind != _smi.SMI_INDEX_INDEX:
-            raise SMIException("child {0} of {1} has an unhandled kind of index".format(
-                ffi.string(child.name),
-                ffi.string(self.node.name)))
+            raise SMIException("child {0} of {1} has an unhandled "
+                               "kind of index".format(
+                                   ffi.string(child.name),
+                                   ffi.string(self.node.name)))
         return child
 
     @property
     def implied(self):
-        """Is the last index implied?
+        """Is the last index implied? An implied index is an index whose size
+        is not fixed but who is not prefixed by its size because this
+        is the last index of a table.
 
-        @return: True iff this is the case
+        :return: `True` if and only if the last index is implied.
         """
         child = self._row
         if child.implied:
@@ -350,9 +402,11 @@ class Table(Entity):
 
     @property
     def index(self):
-        """Get indexes for a table.
+        """Get indexes for a table. The indexes are used to locate a precise
+        row in a table. They are a subset of the table columns.
 
-        @return: list of indexes (as columns) of the table
+        :return: The list of indexes (as :class:`Column` instances) of
+            the table.
         """
         child = self._row
         lindex = []
@@ -360,24 +414,29 @@ class Table(Entity):
         while element != ffi.NULL:
             nelement = _smi.smiGetElementNode(element)
             if nelement == ffi.NULL:
-                raise SMIException("cannot get index associated with {0}".format(
-                    ffi.string(self.node.name)))
+                raise SMIException("cannot get index "
+                                   "associated with {0}".format(
+                                       ffi.string(self.node.name)))
             if nelement.nodekind != _smi.SMI_NODEKIND_COLUMN:
-                raise SMIException("index {0} for {1} is not a column".format(
-                    ffi.string(nelement.name),
-                    ffi.string(self.node.name)))
+                raise SMIException("index {0} for {1} is "
+                                   "not a column".format(
+                                       ffi.string(nelement.name),
+                                       ffi.string(self.node.name)))
             lindex.append(Column(nelement))
             element = _smi.smiGetNextElement(element)
         return lindex
 
-class Column(Entity):
-    """MIB column entity"""
+
+class Column(Node):
+
+    """MIB column node. This class represent a column of a table."""
 
     @property
     def table(self):
         """Get table associated with this column.
 
-        @return: a Table object
+        :return: The :class:`Table` instance associated to this
+            column.
         """
         parent = _smi.smiGetParentNode(self.node)
         if parent == ffi.NULL:
@@ -398,23 +457,27 @@ class Column(Entity):
         t = Table(parent)
         return t
 
-class Node(Entity):
-    """MIB node entity"""
 
 def reset():
-    """Reset libsmi to its initial state"""
-    _smi.smiExit();
-    if _smi.smiInit(b"snimpy") != 0:
+    """Reset libsmi to its initial state."""
+    _smi.smiExit()
+    if _smi.smiInit(b"snimpy") < 0:
             raise SMIException("unable to init libsmi")
-    _smi.smiSetErrorLevel(0);
-    _smi.smiSetFlags(_smi.SMI_FLAG_ERRORS | _smi.SMI_FLAG_RECURSIVE);
+    _smi.smiSetErrorLevel(0)
+    try:
+        _smi.smiSetFlags(_smi.SMI_FLAG_ERRORS | _smi.SMI_FLAG_RECURSIVE)
+    except TypeError:
+        pass                    # We are being mocked
+
 
 def _get_module(name):
     """Get the SMI module from its name.
 
-    @param name: name of the module
-    @return: SMI module or None if not found (not loaded)
+    :param name: The name of the module
+    :return: The SMI module or `None` if not found (not loaded)
     """
+    if not isinstance(name, bytes):
+        name = name.encode("ascii")
     m = _smi.smiGetModule(name)
     if m == ffi.NULL:
         return None
@@ -422,41 +485,47 @@ def _get_module(name):
         return None
     return m
 
+
 def _kind2object(kind):
     return {
         _smi.SMI_NODEKIND_NODE: Node,
         _smi.SMI_NODEKIND_SCALAR: Scalar,
         _smi.SMI_NODEKIND_TABLE: Table,
         _smi.SMI_NODEKIND_COLUMN: Column
-    }.get(kind, Entity)
+    }.get(kind, Node)
+
 
 def get(mib, name):
     """Get a node by its name.
 
-    @param mib: MIB name to query
-    @param name: object name to get from the MIB
-    @return: the requested MIB entity
+    :param mib: The MIB name to query
+    :param name: The object name to get from the MIB
+    :return: the requested MIB node (:class:`Node`)
     """
+    if not isinstance(mib, bytes):
+        mib = mib.encode("ascii")
     module = _get_module(mib)
     if module is None:
         raise SMIException("no module named {0}".format(mib))
-    node = _smi.smiGetNode(module, name)
+    node = _smi.smiGetNode(module, name.encode("ascii"))
     if node == ffi.NULL:
         raise SMIException("in {0}, no node named {1}".format(
-            mib,name))
+            mib, name))
     pnode = _kind2object(node.nodekind)
     return pnode(node)
 
+
 def _get_kind(mib, kind):
     """Get nodes of a given kind from a MIB.
 
-    @param mib: MIB name to search objects for
-    @param kind: SMI kind of object
-    @return: list of matched MIB entities for the MIB
+    :param mib: The MIB name to search objects for
+    :param kind: The SMI kind of object
+    :return: The list of matched MIB nodes for the MIB
     """
+    if not isinstance(mib, bytes):
+        mib = mib.encode("ascii")
     module = _get_module(mib)
     if module is None:
-        print(mib)
         raise SMIException("no module named {0}".format(mib))
     lnode = []
     node = _smi.smiGetFirstNode(module, kind)
@@ -465,51 +534,63 @@ def _get_kind(mib, kind):
         node = _smi.smiGetNextNode(node, kind)
     return lnode
 
+
 def getNodes(mib):
     """Return all nodes from a given MIB.
 
-    @param mib: MIB name
-    @return: list of all MIB entities for the MIB
+    :param mib: The MIB name
+    :return: The list of all MIB nodes for the MIB
+    :rtype: list of :class:`Node` instances
     """
     return _get_kind(mib, _smi.SMI_NODEKIND_NODE)
 
+
 def getScalars(mib):
     """Return all scalars from a given MIB.
 
-    @param mib: MIB name
-    @return: list of all scalars for the MIB
+    :param mib: The MIB name
+    :return: The list of all scalars for the MIB
+    :rtype: list of :class:`Scalar` instances
     """
     return _get_kind(mib, _smi.SMI_NODEKIND_SCALAR)
 
+
 def getTables(mib):
     """Return all tables from a given MIB.
 
-    @param mib: MIB name
-    @return: list of all tables for the MIB
+    :param mib: The MIB name
+    :return: The list of all tables for the MIB
+    :rtype: list of :class:`Table` instances
     """
     return _get_kind(mib, _smi.SMI_NODEKIND_TABLE)
 
+
 def getColumns(mib):
     """Return all columns from a givem MIB.
 
-    @param mib: MIB name
-    @return: list of all columns for the MIB
+    :param mib: The MIB name
+    :return: The list of all columns for the MIB
+    :rtype: list of :class:`Column` instances
     """
     return _get_kind(mib, _smi.SMI_NODEKIND_COLUMN)
 
+
 def load(mib):
     """Load a MIB into the library.
 
-    @param mib: MIB to load, either a filename or a MIB name
-    @return: MIB name that has been loaded
+    :param mib: The MIB to load, either a filename or a MIB name.
+    :return: The MIB name that has been loaded.
+    :except SMIException: The requested MIB cannot be loaded.
     """
+    if not isinstance(mib, bytes):
+        mib = mib.encode("ascii")
     modulename = _smi.smiLoadModule(mib)
     if modulename == ffi.NULL:
         raise SMIException("unable to load {0}".format(mib))
     modulename = ffi.string(modulename)
-    if not _get_module(modulename):
-        raise SMIException("{0} contains major SMI error (check with smilint -s -l1)".format(
-            mib))
+    if not _get_module(modulename.decode("ascii")):
+        raise SMIException("{0} contains major SMI error "
+                           "(check with smilint -s -l1)".format(mib))
     return modulename
 
 reset()
